#include "SamFile.h"
#include "Pileup.h"
#include "PileupElement.h"
#include "unistd.h"
#include "RangeList.h"
#include "Parameters.h"
#include "Error.h"

typedef std::map< char, int> freqType;

template <class T, class U>
class SiteCoverage: public PileupElement {
public:
    SiteCoverage():depth(0) {
	init();
 };
    SiteCoverage(bool sum) : depth(0){
	init();
	summary = sum;
}
    virtual ~SiteCoverage() {};

    void init(){
    summary = false;
    full = false;
    pass = false;
    min_mapq = 0;
    min_bq = 0;
    min_dp = 0;
    max_dp = -1;

    for(int i=0; i<6; i++) { 
	nt_count[i] = 0; ps_nt_count[i] = 0; ns_nt_count[i] = 0; ps_nt_count_all[i] = 0; ns_nt_count_all[i] = 0; 
	nt_qual[i] = 0; ps_nt_qual[i] = 0; ns_nt_qual[i] = 0; 
	nt_mapq[i] = 0; ps_nt_mapq[i] = 0; ns_nt_mapq[i] = 0; 
	nt_cycle[i] = 0; ps_nt_cycle[i] = 0; ns_nt_cycle[i] = 0; 
        r1f=r1r=r2f=r2r=0;
    }
}

    // Add an entry to this pileup element.
    virtual void addEntry(SamRecord& record) {
        // Call the base class:
        PileupElement::addEntry(record);
       
        int mapq = record.getMapQuality();
	this->alldepth++;
	if(mapq<min_mapq) return;

        // Here is your code:
        Cigar* cigar = record.getCigarInfo();
        int32_t readIndex = 
            cigar->getQueryIndex(getRefPosition(), record.get0BasedPosition());

        if(readIndex != CigarRoller::INDEX_NA)
        {
            char base = record.getSequence(readIndex);
            int qual = record.getQuality(readIndex);
            this->depth++;            
            this->freq[base] ++;
            this->mapq.push_back(mapq);
            this->bases.push_back(base);
            this->cycles.push_back(readIndex+1);
            this->qualities.push_back(qual);
            //this->seq_string.push_back(record.getSequence());
            //this->qual_string.push_back(record.getQuality());
            //this->mapPos.push_back(record.get1BasedPosition());
            //this->cigar_str.push_back((record.getCigar()));
	    this->proper.push_back( (record.getFlag() & 0x2 ) ? true : false );

        if (record.getFlag() & 0x10 ) { // check read strand
        	  if(qual>=min_bq) this->numMinusStrand++;
        	  this->strands.push_back('-');
		  if(record.getFlag() & 0x40) r1r++;
		  else r2r++;
        } else {
            if(qual>=min_bq) this->numPlusStrand++;
            this->strands.push_back('+');
	    if(record.getFlag() & 0x40) r1f++;
  	    else r2f++;
        }

        }
	else {
	//std::cout<<"Warning: invalid read index!\n";
	}
        
    }

    // Perform the alalysis associated with this class.  In this case, it is
    // a print of the base & quality information associated with this position.
    virtual void analyze() {
	if(this->pass && this->depth==0) return; // Only output the results when there is at least one base passed depth the filtering 
        if (getRefPosition() != UNSET_POSITION)
        {
	int    t_nt_count[6] = {0,0,0,0,0,0};
	double t_nt_qual[6] = {0,0,0,0,0,0};
	double t_nt_cycle[6] = {0,0,0,0,0,0};
	double t_nt_mapq[6] = {0,0,0,0,0,0};
	int    t_ps_nt_count[6] = {0,0,0,0,0,0};
	double t_ns_nt_count[6] = {0,0,0,0,0,0};
	double t_ps_nt_qual[6] = {0,0,0,0,0,0};
	double t_ns_nt_qual[6] = {0,0,0,0,0,0};
	double t_ps_nt_cycle[6] = {0,0,0,0,0,0};
	double t_ns_nt_cycle[6] = {0,0,0,0,0,0};
	double t_ps_nt_mapq[6] = {0,0,0,0,0,0};
	double t_ns_nt_mapq[6] = {0,0,0,0,0,0};
	int    nt_proper_count[6] = {0,0,0,0,0,0};
	int    ps_nt_proper_count[6] = {0,0,0,0,0,0};
	int    ns_nt_proper_count[6] = {0,0,0,0,0,0};

	int idx = -1;
	char b;
            for(int i=0; i<this->bases.size(); i++)
	{
		b = toupper(this->bases[i]);
		if(b=='A') idx = 0; if(b=='C')idx=1; if(b=='G')idx=2; if(b=='T')idx=3; if(b=='N')idx=4;
		if(idx<0) continue;
		nt_count[idx]++;
		if(this->proper[i]==true) nt_proper_count[idx]++;
		t_nt_qual[idx]+=this->qualities[i]-33;
		t_nt_cycle[idx]+=abs(this->cycles[i]-50);
		t_nt_mapq[idx]+=this->mapq[i];
		if(this->strands[i]=='+')
		{
		ps_nt_count_all[idx]++;
		if(this->qualities[i]-33<min_bq) continue;
		ps_nt_count[idx]++;
		if(this->proper[i]==true) ps_nt_proper_count[idx]++;
		t_ps_nt_qual[idx]+=this->qualities[i]-33;
		t_ps_nt_cycle[idx]+=abs(this->cycles[i]-50);
		t_ps_nt_mapq[idx]+=this->mapq[i];
		}
		else if(this->strands[i]=='-')
		{
		ns_nt_count_all[idx]++;
		if(this->qualities[i]-33<min_bq) continue;
		ns_nt_count[idx]++;
		if(this->proper[i]==true) ns_nt_proper_count[idx]++;
		t_ns_nt_qual[idx]+=this->qualities[i]-33;
		t_ns_nt_cycle[idx]+=abs(this->cycles[i]-50);
		t_ns_nt_mapq[idx]+=this->mapq[i];
		}
		else error("Strand is not '+' or '-'\n");
		idx = -1;
	}

	    for(int i=0; i<5; i++)
	{
		nt_qual[i] = nt_count[i]==0 ? 0 : t_nt_qual[i]/nt_count[i];
		nt_mapq[i] = nt_count[i]==0 ? 0 : t_nt_mapq[i]/nt_count[i];
		nt_cycle[i] = nt_count[i]==0 ? 0 : t_nt_cycle[i]/nt_count[i];
		ps_nt_qual[i] = ps_nt_count[i]==0 ? 0 : t_ps_nt_qual[i]/ps_nt_count[i];
		ps_nt_mapq[i] = ps_nt_count[i]==0 ? 0 : t_ps_nt_mapq[i]/ps_nt_count[i];
		ps_nt_cycle[i] = ps_nt_count[i]==0 ? 0 : t_ps_nt_cycle[i]/ps_nt_count[i];
		ns_nt_qual[i] = ns_nt_count[i]==0 ? 0 : t_ns_nt_qual[i]/ns_nt_count[i];
		ns_nt_mapq[i] = ns_nt_count[i]==0 ? 0 : t_ns_nt_mapq[i]/ns_nt_count[i];
		ns_nt_cycle[i] = ns_nt_count[i]==0 ? 0 : t_ns_nt_cycle[i]/ns_nt_count[i];
	}

            if(this->alldepth < min_dp) return;
            if(max_dp>0 && this->alldepth > max_dp) return;

            std::cout << getChromosome() << "\t" << getRefPosition()+1<<"\t";
            std::cout << this->alldepth <<"\t"<<this->depth<<"\t"<<this->numPlusStrand<<"\t"<<this->numMinusStrand<<"\t";
	    //std::cout << this->r1f <<":"<< this->r2f << "|" << this->r1r << ":" << this->r2r <<"\t";
	    for(int i=0; i<4; i++) {
	 	std::cout << ps_nt_count_all[i]<< ":" << ns_nt_count_all[i];
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";

	    for(int i=0; i<4; i++) {
	 	std::cout << ps_nt_count[i]<< ":" << ns_nt_count[i];
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";

/*
	    for(int i=0; i<4; i++) {
	 	std::cout << ps_nt_proper_count[i]<< ":" << ns_nt_proper_count[i];
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";
*/

if(summary==false)
{
	    std::setprecision(1);	    

	    for(int i=0; i<4; i++) {
		printf("%.1f:%.1f", ps_nt_qual[i], ns_nt_qual[i]);
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";

	    for(int i=0; i<4; i++) {
	 	printf("%.1f:%.1f", ps_nt_cycle[i],ns_nt_cycle[i]);
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";

	    for(int i=0; i<4; i++) {
	 	printf("%.1f:%.1f", ps_nt_mapq[i],ns_nt_mapq[i]);
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";

	    if(full)
            for(int i=0; i<this->bases.size(); i++)
            {
            	std::cout << int(this->mapq[i])<<this->bases[i]<< this->cycles[i] << this->strands[i] <<this->qualities[i]-33<<"\t"; 
            	//std::cout<<cigar_str[i]<<":"<<seq_string[i]<<":"<<qual_string[i]<<"\t";
            }
}
            std::cout<<"\n";
        }
    };
    void reset(int refPosition) {
        // Call the base class.
        PileupElement::reset(refPosition);
	this->alldepth = 0;
        this->depth = 0;
        this->numMinusStrand = 0;
        this->numPlusStrand = 0;
        this->freq.clear();
        this->base.clear();
        this->bases.clear();
        this->qualities.clear();
        this->cycles.clear();
        this->strands.clear();
        this->seq_string.clear();
        this->qual_string.clear();
        this->cigar_str.clear();
	this->mapq.clear();
	this->proper.clear();
	for(int i=0;i<6;i++){
 	nt_count[i] = 0; ps_nt_count[i]=0; ns_nt_count[i]=0; ps_nt_count_all[i]=0; ns_nt_count_all[i]=0;
	nt_qual[i] = 0; ps_nt_qual[i]=.0;ns_nt_qual[i]=0;
	nt_cycle[i] = 0; ps_nt_cycle[i]=.0;ns_nt_cycle[i]=0;
	nt_mapq[i] = 0; ps_nt_mapq[i]=.0;ns_nt_mapq[i]=0;
	}
	r1f=r1r=r2f=r2r=0;
    };
    
private:
    U depth;
    T freq; // freq of A, C, G, T
    int alldepth;
    int nt_count[6];
    int ps_nt_count[6];
    int ns_nt_count[6];
    int ps_nt_count_all[6];
    int ns_nt_count_all[6];
    double nt_qual[6];
    double ps_nt_qual[6];
    double ns_nt_qual[6];
    double nt_cycle[6];
    double ps_nt_cycle[6];
    double ns_nt_cycle[6];
    double nt_mapq[6];
    double ps_nt_mapq[6];
    double ns_nt_mapq[6];
    int numPlusStrand;
    int numMinusStrand;
    std::string base;
    std::vector<char> mapq;
    std::vector<char> bases;
    std::vector<int> qualities;
    std::vector<int> cycles;
    std::vector<char> strands;
    std::vector<std::string> seq_string;
    std::vector<std::string> qual_string;
    std::vector<unsigned int> mapPos;
    std::vector<std::string> cigar_str;
    std::vector<bool> proper;
    int r1f, r1r, r2f, r2r;
public:
    bool summary, full, pass;
    int min_mapq, min_bq, min_dp, max_dp;
    FILE *OUT;
};

template <class T>
class RegionPileup: public Pileup<T>{

public:
  int min_mapq, min_bq, min_dp, max_dp;
  bool proper_paired_keep, dup_keep, summary, full, pass;
public:
    RegionPileup():Pileup<T>() { min_mapq = 0; min_bq=0; min_dp=0; max_dp=-1;proper_paired_keep = false; dup_keep = false; summary=false; full=false; pass=false; };
    RegionPileup(int window):Pileup<T>(window) {};


void addElement(T& element, SamRecord& record)
{   
    element.min_mapq = min_mapq;
    element.min_bq = min_bq;
    element.min_dp = min_dp;
    element.max_dp = max_dp;
    element.summary = summary;
    element.full = full;
    element.pass = pass;
    element.addEntry(record);
}

    virtual int processFile(const std::string& fileName, 
                            uint16_t excludeFlag = 0x0704, 
                            uint16_t includeFlag = 0) {
        SamFile samIn;
        SamFileHeader header;
        SamRecord record;
    
        // if(myRefPtr != NULL)
        // {
        //     samIn.SetReference(myRefPtr);
        // }

        if(!samIn.OpenForRead(fileName.c_str()))
        {
            fprintf(stderr, "%s\n", samIn.GetStatusMessage());
            return(samIn.GetStatus());
        }
    
        if(!samIn.ReadHeader(header))
        {
            fprintf(stderr, "%s\n", samIn.GetStatusMessage());
            return(samIn.GetStatus());
        }

        // The file needs to be sorted by coordinate.
        samIn.setSortedValidation(SamFile::COORDINATE);

        // Iterate over all records
        while (samIn.ReadRecord(header, record))
        {
	    //if(record.getMapQuality() < min_mapq) continue; // filtering will be done during pileup to count all mapped bases
            uint16_t flag = record.getFlag();
            if(flag & 0x4 || flag & 0x200) continue;
	    if((flag & 0x2)==0 && proper_paired_keep) continue;
	    if(flag & 0x400) if(dup_keep==false) continue;

            if(flag & excludeFlag)
            {
                // This record has an excluded flag set, 
                // so continue to the next one.
                continue;
            }
            if((flag & includeFlag) != includeFlag)
            {
                // This record does not have all required flags set, 
                // so continue to the next one.
                continue;
            }
            Pileup<T>::processAlignment(record);
        }

        Pileup<T>::flushPileup();

        int returnValue = 0;
        if(samIn.GetStatus() != SamStatus::NO_MORE_RECS)
        {
            // Failed to read a record.
            fprintf(stderr, "%s\n", samIn.GetStatusMessage());
            returnValue = samIn.GetStatus();
        }
        return(returnValue);  
    };
    int processRegion(const char*  bamFileName,
                      const char*  refName,
                      int start,
                      int end) {
        std::string _indexFileName;
        _indexFileName = bamFileName;
        _indexFileName += ".bai";
        int ret = processRegion(bamFileName,
                             _indexFileName.c_str(),
                             refName,
                             start,
                             end);
        Pileup<T>::flushPileup();
        return ret;
    };
    int processRegion(const char*  bamFileName,
                      const char*  indexFileName,
                      const char*  refName,
                      int start,
                      int end) {
        SamFile samIn;
        SamFileHeader header;
        SamRecord record;
        if(!samIn.OpenForRead(bamFileName))
        {
            fprintf(stderr, "%s\n", samIn.GetStatusMessage());
            return(samIn.GetStatus());
        }

        if (!samIn.SetReadSection(refName, start, end))
        {
            fprintf(stderr, "%s\n", samIn.GetStatusMessage());
            return(samIn.GetStatus());
        }
        
        // Open the bam index file for reading.
        if (!samIn.ReadBamIndex(indexFileName)) {
            fprintf(stderr, "%s\n", samIn.GetStatusMessage());
            return(samIn.GetStatus());
        };

        if(!samIn.ReadHeader(header))
        {
            fprintf(stderr, "%s\n", samIn.GetStatusMessage());
            return(samIn.GetStatus());
        }
        int numSectionRecords;
        // Set returnStatus to success.  It will be changed
        // to the failure reason if any of the writes fail.
        SamStatus::Status returnStatus = SamStatus::SUCCESS;

        while(samIn.ReadRecord(header, record))
        {
	    //if(record.getMapQuality() < min_mapq ) continue;
	    uint16_t flag = record.getFlag();
            if(flag & 0x4 || flag&0x200) continue;
	    if((flag & 0x2)==0 && proper_paired_keep) continue;
	    if(flag & 0x400) if(dup_keep==false) continue;

            // Successfully read a record from the file, so pile it up.
            this->processAlignment(record, start, end);
            ++numSectionRecords;
        }
        return (returnStatus);
    };
    virtual void processAlignment(SamRecord& record, 
                                  const int begin, 
                                  const int end) {
        int refPosition = record.get0BasedPosition();
        int refID = record.getReferenceID();

        // Flush any elements from the pileup that are prior to this record
        // since the file is sorted, we are done with those positions.
        Pileup<T>::flushPileup(refID, refPosition);
    
        // Loop through for each reference position covered by the record.
        // It is up to the PILEUP_TYPE to handle insertions/deletions, etc
        // that are related with the given reference position.
        for(; refPosition <= record.get0BasedAlignmentEnd(); ++refPosition)
        {
            if (begin <= refPosition && refPosition < end)
                // check refPosition is in specified region
                Pileup<T>::addAlignmentPosition(refPosition, record);
        }
    }
};

void usage(int argc, char *argv[]) {
    std::cout << argv[0] << " a.bam -- pile up a.bam" << std::endl;
    std::cout << argv[0] << " a.bam rangeListFile -- pile up given region of a.bam" << std::endl;
};

#define PILEUP_WINDOW_SIZE 1024

int main(int argc, char *argv[])
{
/*
    if (argc != 2 && argc != 3) {
        usage(argc, argv);
        exit(2);
    }
*/
String sites;
String outFile;
int minMapQual = 0;
int minBaseQual = 0;
int minDepth = 0;
int maxDepth = -1;

StringArray bamFiles;
String bam;
bool proper_keep = false;
bool dup_keep = false;
bool summary = false;
bool full = false;
bool pass = false;

  ParameterList pl;

  BEGIN_LONG_PARAMETERS(longParameters)
    LONG_PARAMETER_GROUP("Input Files")
      LONG_STRINGPARAMETER("bam",&bam)
      LONG_STRINGPARAMETER("bed",&sites)
    LONG_PARAMETER_GROUP("Filters")
      LONG_INTPARAMETER("min_mapq",&minMapQual)
      LONG_INTPARAMETER("min_bq",&minBaseQual)
      LONG_INTPARAMETER("min_dp",&minDepth)
      LONG_INTPARAMETER("max_dp",&maxDepth)
      LONG_PARAMETER("proper_keep",&proper_keep)
      LONG_PARAMETER("dup_keep",&dup_keep)
    LONG_PARAMETER_GROUP("Output")
      //LONG_STRINGPARAMETER("out",&outFile)
      LONG_PARAMETER("summary",&summary)
      LONG_PARAMETER("full",&full)
      LONG_PARAMETER("pass",&pass)
   END_LONG_PARAMETERS();

  pl.Add(new LongParameters("", longParameters));
  pl.Read(argc, argv);
  /*
  int in = pl.ReadWithTrailer(argc, argv, 1);
  for (int i=in+1; i<argc; i++){
        bamFiles.Push(argv[i]);
  }
 */
  pl.Status();

  if(bam.Length()==0) error("No bam file provided!\n");
//  if(outFile.Length()==0) error("No output file provided!\n");

//  FILE *OUT = fopen(outFile.c_str(), "w");
//  if(OUT==NULL) {fprintf(stderr, "Open file for output failed!\n"; exit(1); }

    //printf("processing %s...\n", argv[1]);
    if (sites.Length()>0) {
	std::string s(bam.c_str());
        RangeList rl;
        rl.addRangeFile(sites.c_str());
        //rl.addRangeFile(sites);
        std::string chr;
        int begin;
        int end;

        for (unsigned int i = 0; i < rl.size(); i++){
            rl.obtainRange(i, &chr, &begin, &end);
	    if(chr.compare("")==0 || begin<=0 || end<begin) error("Regions are not correctly specified!\n");
            RegionPileup<SiteCoverage<freqType, int > > pileup(PILEUP_WINDOW_SIZE);
            //printf("rangelist %d: chr %s %d - %d\n", i, chr.c_str(), begin, end);
	    pileup.min_mapq = minMapQual;
	    pileup.min_bq = minBaseQual;
	    pileup.min_dp = minDepth;
	    pileup.max_dp = maxDepth;
	    pileup.proper_paired_keep = proper_keep;
	    pileup.dup_keep = dup_keep;
	    pileup.summary = summary;
	    pileup.full = full;
	    pileup.pass = pass;
            pileup.processRegion(bam.c_str(), chr.c_str(), begin-1, end);
        }
    }
    if (sites.Length()==0) {
	error("A list of regions need be provided!\n");
	std::string s(bam.c_str());
        Pileup<SiteCoverage<freqType, int > > pileup(PILEUP_WINDOW_SIZE);
        pileup.processFile(s);
    }
    return 0;
}
